# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).


from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt

from kqcircuits.chips.chip import Chip
from kqcircuits.defaults import default_airbridge_type, default_sampleholders
from kqcircuits.elements.waveguide_coplanar import WaveguideCoplanar
from kqcircuits.elements.quarter_wave_cpw_resonator import QuarterWaveCpwResonator


class QualityFactor(Chip):
    """The PCell declaration for a QualityFactor chip."""

    res_lengths = Param(
        pdt.TypeList,
        "Resonator lengths",
        [5434, 5429, 5374, 5412, 5493, 5589],
        unit="[μm]",
        docstring="Physical length of resonators [μm]",
    )
    n_fingers = Param(
        pdt.TypeList, "Number of fingers of the coupler", [4, 4, 2, 4, 4, 4], docstring="Fingers in planar capacitors"
    )
    l_fingers = Param(
        pdt.TypeList,
        "Length of fingers",
        [23.1, 9.9, 14.1, 10, 21, 28],
        unit="[μm]",
        docstring="Length of the capacitor fingers [μm]",
    )
    type_coupler = Param(
        pdt.TypeList, "Coupler types", ["interdigital", "interdigital", "interdigital", "gap", "gap", "gap"]
    )
    n_ab = Param(pdt.TypeList, "Number of resonator airbridges", [5, 0, 5, 5, 5, 5])
    res_term = Param(
        pdt.TypeList,
        "Resonator termination type",
        ["galvanic", "galvanic", "galvanic", "airbridge", "airbridge", "airbridge"],
    )
    res_beg = Param(
        pdt.TypeList,
        "Resonator beginning type",
        ["galvanic", "galvanic", "galvanic", "airbridge", "airbridge", "airbridge"],
    )
    res_a = Param(
        pdt.TypeList,
        "Resonator waveguide center conductor width",
        [5, 10, 20, 5, 10, 20],
        unit="[μm]",
        docstring="Width of the center conductor in the resonators [μm]",
    )
    res_b = Param(
        pdt.TypeList,
        "Resonator waveguide gap width",
        [3, 6, 12, 3, 6, 12],
        unit="[μm]",
        docstring="Width of the gap in the resonators [μm]",
    )
    tl_airbridges = Param(pdt.TypeBoolean, "Airbridges on transmission line", True)
    res_airbridge_types = Param(pdt.TypeList, "Airbridge type for each resonator", default=[default_airbridge_type] * 6)
    sample_holder_type = Param(pdt.TypeInt, "Sample holder type for the chip", "SMA8", choices=["SMA8", "ARD24"])
    marker_safety = Param(pdt.TypeDouble, "Distance between launcher and first curve", 1000, unit="μm")
    feedline_bend_distance = Param(pdt.TypeDouble, "Horizontal distance of feedline bend", 100, unit="μm")
    resonators_both_sides = Param(pdt.TypeBoolean, "Place resonators on both sides of feedline", False)
    max_res_len = Param(
        pdt.TypeDouble,
        "Maximal straight length of resonators",
        1e30,
        unit="μm",
        docstring="Resonators exceeding this length become meandering",
    )
    ground_grid_in_trace = Param(pdt.TypeList, "Include ground-grid in the trace", [0] * 18)
    # override box to have hidden=False and allow GUI editing
    box = Param(pdt.TypeShape, "Border", pya.DBox(pya.DPoint(0, 0), pya.DPoint(10000, 10000)))

    def build(self):
        # Interpretation of parameter lists
        res_lengths = [float(foo) for foo in self.res_lengths]
        res_a = [float(foo) for foo in self.res_a]
        res_b = [float(foo) for foo in self.res_b]
        n_fingers = [float(foo) for foo in self.n_fingers]
        type_coupler = self.type_coupler
        n_ab = [int(foo) for foo in self.n_ab]
        l_fingers = [float(foo) for foo in self.l_fingers]
        res_term = self.res_term
        res_beg = self.res_beg

        # center the resonators in the chip regardless of size
        max_res_len = min(max(res_lengths), self.max_res_len)
        chip_side = self.box.p2.y - self.box.p1.y
        if self.resonators_both_sides:
            wg_top_y = chip_side / 2
        else:
            wg_top_y = (chip_side + max_res_len) / 2

        # support resizable chip keeping pad distances from the top constant
        if self.sample_holder_type == "ARD24":
            launchers = self.produce_n_launchers(
                **{**default_sampleholders["ARD24"], "pad_pitch": (chip_side - 4000) / 5, "chip_box": self.box},
                launcher_assignments={24: "PL-1-IN", 7: "PL-1-OUT"},
            )
        elif self.sample_holder_type == "SMA8":
            launchers = self.produce_n_launchers(
                **{**default_sampleholders["SMA8"], "pad_pitch": chip_side - 2 * 2800, "chip_box": self.box},
                launcher_assignments={8: "PL-1-IN", 3: "PL-1-OUT"},
            )

        # Define start and end of feedline
        points_fl = [launchers["PL-1-IN"][0]]
        if abs(launchers["PL-1-IN"][0].y - wg_top_y) > 1:
            # Bend in the feedline needed
            points_fl += [
                launchers["PL-1-IN"][0] + pya.DVector(self.r + self.marker_safety, 0),
                pya.DPoint(
                    launchers["PL-1-IN"][0].x + self.r + self.feedline_bend_distance + self.marker_safety, wg_top_y
                ),
            ]
            points_fl_end = [
                pya.DPoint(
                    launchers["PL-1-OUT"][0].x - self.r - self.feedline_bend_distance - self.marker_safety, wg_top_y
                ),
                launchers["PL-1-OUT"][0] + pya.DVector(-self.r - self.marker_safety, 0),
            ]
        elif self.marker_safety > 0:
            points_fl += [launchers["PL-1-IN"][0] + pya.DVector(self.marker_safety, 0)]
            points_fl_end = [
                launchers["PL-1-OUT"][0] + pya.DVector(-self.marker_safety, 0),
            ]
        else:
            points_fl_end = []

        points_fl_end += [launchers["PL-1-OUT"][0]]

        tl_start = points_fl[-1]
        tl_end = points_fl_end[0]

        resonators = len(self.res_lengths)
        v_res_step = (tl_end - tl_start) * (1.0 / resonators)

        for i in range(resonators):
            resonator_up = self.resonators_both_sides and (i % 2) == 0

            # Add resonator element
            _, refp = self.insert_cell(
                QuarterWaveCpwResonator,
                trans=pya.DTrans(2 if resonator_up else 0, False, tl_start + (i + 0.5) * v_res_step),
                probeline_len=120.0,
                resonator_length=res_lengths[i],
                max_res_len=max_res_len,
                res_beg=res_beg[i],
                res_term=res_term[i],
                n_ab=n_ab[i],
                res_airbridge_types=str(self.res_airbridge_types[i]),
                tl_airbridges=self.tl_airbridges,
                n_fingers=n_fingers[i],
                l_fingers=l_fingers[i],
                ground_grid_in_trace=int(self.ground_grid_in_trace[i]),
                type_coupler=type_coupler[i],
                res_a=res_a[i],
                res_b=res_b[i],
            )

            # Add segment of feedline
            self.insert_cell(
                WaveguideCoplanar,
                **{
                    **self.cell.pcell_parameters_by_name(),
                    **{
                        "path": pya.DPath(points_fl + [refp["port_b" if resonator_up else "port_a"]], 1),
                        "term2": 0,
                        "ground_grid_in_trace": False,
                    },
                },
            )
            points_fl = [refp["port_a" if resonator_up else "port_b"]]

        # Add the last segment of feedline
        self.insert_cell(
            WaveguideCoplanar,
            **{
                **self.cell.pcell_parameters_by_name(),
                **{"path": pya.DPath(points_fl + points_fl_end, 1), "term2": 0, "ground_grid_in_trace": False},
            },
        )



# This code is part of KQCircuits
# Copyright (C) 2024 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).

from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt
from kqcircuits.elements.element import Element
from kqcircuits.elements.waveguide_coplanar import WaveguideCoplanar
from kqcircuits.elements.waveguide_coplanar_splitter import WaveguideCoplanarSplitter, t_cross_parameters
from kqcircuits.util.coupler_lib import cap_params
from kqcircuits.util.refpoints import WaveguideToSimPort
from kqcircuits.elements.waveguide_composite import WaveguideComposite, Node
from kqcircuits.elements.airbridges.airbridge import Airbridge
from kqcircuits.elements.airbridge_connection import AirbridgeConnection
from kqcircuits.defaults import default_airbridge_type
from kqcircuits.elements.airbridges import airbridge_type_choices


class QuarterWaveCpwResonator(Element):
    """
    CPW Resonator
    """

    probeline_length = Param(pdt.TypeDouble, "Probeline length", 500, unit="μm")
    resonator_length = Param(pdt.TypeDouble, "Resonator length ", 1000, unit="μm")
    max_res_len = Param(
        pdt.TypeDouble,
        "Maximal straight length of resonators",
        1e30,
        unit="μm",
        docstring="Resonators exceeding this length become meandering",
    )
    res_beg = Param(pdt.TypeString, "Resonator beginning type", "galvanic")
    res_term = Param(pdt.TypeString, "Resonator termination type", "galvanic")
    n_ab = Param(pdt.TypeInt, "Number of resonator airbridges", 0)
    res_airbridge_type = Param(
        pdt.TypeString, "Airbridge type", default=default_airbridge_type, choices=airbridge_type_choices
    )
    tl_airbridges = Param(pdt.TypeBoolean, "Airbridges on transmission line", True)
    n_fingers = Param(pdt.TypeDouble, "Number of fingers", 1, unit="")
    l_fingers = Param(pdt.TypeDouble, "Finger length", 1, unit="")
    ground_grid_in_trace = Param(pdt.TypeBoolean, "Include ground-grid in the trace", False)

    type_coupler = Param(
        pdt.TypeString,
        "Capacitor coupler type",
        "smooth",
        choices=[
            ["Interdigital", "interdigital"],
            ["Gap", "gap"],
            ["Smooth", "smooth"],
            ["Ground gap", "ground gap"],
        ],
    )

    res_a = Param(pdt.TypeDouble, "Trace width of resonator line", 10, unit="μm")
    res_b = Param(pdt.TypeDouble, "Gap width of resonator line", 10, unit="μm")

    use_internal_ports = Param(pdt.TypeBoolean, "Internal ports (EPR)", False)

    def build(self):

        cell_cross = self.add_element(
            WaveguideCoplanarSplitter,
            **t_cross_parameters(
                length_extra_side=2 * self.a,
                a=self.a,
                b=self.b,
                a2=self.a,
                b2=self.b,
            ),
        )

        cell_ab_crossing = self.add_element(Airbridge)

        # Cross
        _, cross_refpoints_abs = self.insert_cell(cell_cross)

        points_pl1 = [pya.DPoint(-self.probeline_length / 2.0, 0)]
        points_pl1 += [cross_refpoints_abs["port_left"]]

        points_pl2 = [cross_refpoints_abs["port_right"]]
        points_pl2 += [pya.DPoint(self.probeline_length / 2.0, 0)]

        # Coupler
        _, cplr_refpoints_abs = self.insert_cell(
            trans=pya.DTrans.R90,
            align="port_b",
            align_to=cross_refpoints_abs["port_bottom"],
            **cap_params(
                self.n_fingers,
                self.l_fingers,
                self.type_coupler,
                element_key="cell",
                a=self.res_a,
                b=self.res_b,
                a2=self.a,
                b2=self.b,
            ),
        )

        pos_res_start = cplr_refpoints_abs["port_a"]
        pos_res_end = pos_res_start - pya.DVector(0, min(self.resonator_length, self.max_res_len))

        # create resonator using WaveguideComposite
        if self.res_beg == "airbridge":
            node_beg = Node(pos_res_start, AirbridgeConnection, with_side_airbridges=False)
        else:
            node_beg = Node(pos_res_start)

        length_increment = (
            self.resonator_length - self.max_res_len if self.resonator_length > self.max_res_len else None
        )

        bridge_approach = 38.0
        bridge_length = self.res_a + 2 * self.res_b + bridge_approach

        if self.res_term == "airbridge":
            node_end = Node(
                pos_res_end,
                AirbridgeConnection,
                with_side_airbridges=False,
                with_right_waveguide=False,
                n_bridges=self.n_ab,
                bridge_length=bridge_length,
                length_increment=length_increment,
            )
        else:
            node_end = Node(
                pos_res_end, n_bridges=self.n_ab, bridge_length=bridge_length, length_increment=length_increment
            )

        airbridge_type = self.res_airbridge_type
        resonator_element = self.add_element(
            WaveguideComposite,
            nodes=[node_beg, node_end],
            a=self.res_a,
            b=self.res_b,
            ground_grid_in_trace=self.ground_grid_in_trace,
            airbridge_type=airbridge_type,
        )

        cells_pl1, _ = self.insert_cell(WaveguideCoplanar, path=points_pl1)
        cells_pl2, _ = self.insert_cell(WaveguideCoplanar, path=points_pl2)
        cells_resonator, _ = self.insert_cell(resonator_element)

        # airbridges on the left and right side of the couplers
        if self.tl_airbridges:
            ab_dist_to_coupler = 60.0
            ab_coupler_left = pya.DPoint(
                (cross_refpoints_abs["port_left"].x) - ab_dist_to_coupler, (cross_refpoints_abs["port_left"].y)
            )
            ab_coupler_right = pya.DPoint(
                (cross_refpoints_abs["port_right"].x) + ab_dist_to_coupler, (cross_refpoints_abs["port_right"].y)
            )

            self.insert_cell(cell_ab_crossing, pya.DTrans(0, False, ab_coupler_left))
            self.insert_cell(cell_ab_crossing, pya.DTrans(0, False, ab_coupler_right))

        self.copy_port("a", cells_pl1)
        self.copy_port("b", cells_pl2)

        self.copy_port("a", cells_pl1, "sim_a")
        self.copy_port("b", cells_pl2, "sim_b")
        self.copy_port("b", cells_resonator, "resonator_b")

    @classmethod
    def get_sim_ports(cls, simulation):
        return [
            WaveguideToSimPort(
                "port_sim_a", use_internal_ports=simulation.use_internal_ports, a=simulation.a, b=simulation.b
            ),
            WaveguideToSimPort(
                "port_sim_b", use_internal_ports=simulation.use_internal_ports, a=simulation.a, b=simulation.b
            ),
            WaveguideToSimPort("port_resonator_b", a=simulation.res_a, b=simulation.res_b),
        ]



# This code is part of KQCircuits
# Copyright (C) 2024 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).

import logging
from functools import lru_cache
from pathlib import Path
from kqcircuits.pya_resolver import pya


# Expected properties of the text geometry as loaded from the ``font_polygons.oas`` file

# File path to the ``font_polygons.oas`` file
OAS_PATH = str(Path(__file__).parent.joinpath("font_polygons.oas"))
# Letters in the oas file
OAS_LETTERS = [chr(ord("A") + i) for i in range(0, 32)] + [chr(ord("0") + i) for i in range(0, 10)]
# Spacing between two letters
OAS_TEXT_SPACING = 300.0
# "mag" parameter set for TEXT pcell in oas file
OAS_TEXT_MAGNIFICATION = 500.0
# dbu in the oas file
OAS_DBU = 0.001


def get_text_polygon(label: str, size: int = OAS_TEXT_MAGNIFICATION) -> pya.Region:
    """Returns the given label string as a region.

    If size argument is given as non-integer, will round it to nearest micron.

    Utilizes speed ups compared to generating text geometry with KLayout's TEXT PCells.
    Only supports characters layed out in the ``font_polygons.oas`` file.
    """

    font_polygons = load_font_polygons()
    unsported_characters = set(x.upper() for x in label) - set(font_polygons.keys()) - set(" ")
    if unsported_characters:
        logging.warning(
            f"Unsupported characters for get_text_polygon: {unsported_characters}."
            f" These characters will be skipped in label '{label}'"
        )

    spacing = size * OAS_TEXT_SPACING / (OAS_DBU * OAS_TEXT_MAGNIFICATION)
    label_region = pya.Region()
    if label is not None:
        for i, letter in enumerate(str(label)):
            if letter.upper() not in font_polygons:
                continue
            label_region += (
                font_polygons[letter.upper()]
                .scaled_and_snapped(0, round(size), OAS_TEXT_MAGNIFICATION, 0, round(size), OAS_TEXT_MAGNIFICATION)
                .moved(i * spacing, 0)
            )
    return label_region


@lru_cache(maxsize=None)
def load_font_polygons() -> dict[str, pya.Region]:
    """Loads from static OAS file a region for each letter used in labels.

    Cached for reuse.
    """
    layout = pya.Layout()
    layout.read(OAS_PATH)

    font_dict = {letter: pya.Region() for letter in OAS_LETTERS}
    for shape in pya.Region(layout.top_cells()[-1].begin_shapes_rec(layout.layer(129, 1))).each():
        index = round(shape.bbox().to_dtype(OAS_DBU).p1.x) // OAS_TEXT_MAGNIFICATION
        font_dict[OAS_LETTERS[int(index)]] += pya.Region(shape.moved(-OAS_TEXT_MAGNIFICATION * index / OAS_DBU, 0))

    return font_dict
